"""
Tests for ``nitrate.comments``.
"""

import pytest

from nitrate.comments import fix_comment_text


@pytest.mark.parametrize(
    "comment_text",
    [
        "",
        "Love old gravestones",
        "Text with\na newline",
        'This is a link to <a href="https://example.com" rel="noreferrer nofollow">example.com</a>.',
        'Link to <a href="https://example.com" rel="noreferrer nofollow">example.com</a> in a sentence',
    ],
)
def test_text_is_unchanged(comment_text: str) -> None:
    """
    All of these examples are left unchanged by the comment fixes.
    """
    assert fix_comment_text(comment_text) == comment_text


class TestFixWikipediaLinks:
    """
    Tests for the Wikipedia link fixes.
    """

    @pytest.mark.vcr()
    def test_fixes_trailing_period(self) -> None:
        """
        If there's a Wikipedia URL that ends with a period, that period
        is correctly identified as part of the URL.

        See https://github.com/Flickr-Foundation/commons.flickr.org/issues/297
        """
        comment_text = (
            '<a href="https://en.wikipedia.org/wiki/William_Barnes_Jr" '
            'rel="noreferrer nofollow">en.wikipedia.org/wiki/William_Barnes_Jr</a>.'
        )
        expected_text = (
            '<a href="https://en.wikipedia.org/wiki/William_Barnes_Jr." '
            'rel="noreferrer nofollow">en.wikipedia.org/wiki/William_Barnes_Jr.</a>'
        )

        assert fix_comment_text(comment_text) == expected_text

    @pytest.mark.vcr()
    def test_fixes_disambiguation_parens(self) -> None:
        """
        If there's a Wikipedia URL that has disambiguation parens at
        the end, that extra text is identified as part of the URL.
        """
        # This example comes from Jessamyn's comment here:
        # https://www.flickr.com/photos/twm_news/5257092205/#comment72157720409554465
        #
        # Retrieved 21 January 2025
        comment_text = 'This guy! <a href="https://en.wikipedia.org/wiki/Reg_Dixon_" rel="noreferrer nofollow">en.wikipedia.org/wiki/Reg_Dixon_</a>(comedian)'
        expected_text = 'This guy! <a href="https://en.wikipedia.org/wiki/Reg_Dixon_(comedian)" rel="noreferrer nofollow">en.wikipedia.org/wiki/Reg_Dixon_(comedian)</a>'

        assert fix_comment_text(comment_text) == expected_text

    @pytest.mark.vcr()
    def test_fixes_mobile_wikipedia_links(self) -> None:
        """
        If there's a mobile Wikipedia link that has disambiguation parens
        at the end, that extra text is added to the URL.
        """
        # This example comes from Barbara Agnew's comment here:
        # https://www.flickr.com/photos/royalaustralianhistoricalsociety/12643854003/#comment72157643236765583
        #
        # Retrieved 22 January 2025
        #
        # Note: we need to preserve both the original scheme (http://)
        # and hostname (en.m.wikipedia.org) because we don't want to
        # be doing any rewriting of people's comments.
        comment_text = '<a href="http://en.m.wikipedia.org/wiki/Black_Thursday_" rel="nofollow">en.m.wikipedia.org/wiki/Black_Thursday_</a>(1851)'
        expected_text = '<a href="http://en.m.wikipedia.org/wiki/Black_Thursday_(1851)" rel="noreferrer nofollow">en.m.wikipedia.org/wiki/Black_Thursday_(1851)</a>'

        assert fix_comment_text(comment_text) == expected_text

    @pytest.mark.vcr()
    def test_fixes_dutch_wikipedia_links(self) -> None:
        """
        If there's a non-English Wikipedia link that has disambiguation parens
        at the end, that extra text is added to the URL.
        """
        # This example comes from Ronald van Ooijen's comment:
        # https://www.flickr.com/photos/statelibraryofnsw/49868448366/#comment72157714245484787
        #
        # Retrieved 22 January 2025
        comment_text = (
            "Ship was until 1934 in use and afterwards scrapped.\n"
            "see for total history\n"
            '<a href="https://nl.wikipedia.org/wiki/Mauretania_" rel="noreferrer nofollow">nl.wikipedia.org/wiki/Mauretania_</a>(schip,_1907)'
        )
        expected_text = (
            "Ship was until 1934 in use and afterwards scrapped.\n"
            "see for total history\n"
            '<a href="https://nl.wikipedia.org/wiki/Mauretania_(schip,_1907)" rel="noreferrer nofollow">nl.wikipedia.org/wiki/Mauretania_(schip,_1907)</a>'
        )

        assert fix_comment_text(comment_text) == expected_text

    @pytest.mark.vcr()
    def test_includes_fragment(self) -> None:
        """
        If there's a Wikipedia URL with a fragment, that fragment is added
        to the URL.
        """
        # This example comes from Jesús Roberto Duarte's comment:
        # https://www.flickr.com/photos/cornelluniversitylibrary/3675113899/#comment72157631107864396
        #
        # Retrieved 22 January 2025
        comment_text = (
            "In\n"
            '<a href="http://es.wikipedia.org/wiki/Atoyac_" rel="nofollow">es.wikipedia.org/wiki/Atoyac_</a>(Veracruz)#Toponimia\n'
            "you'll find its coordinates"
        )
        expected_text = (
            "In\n"
            '<a href="http://es.wikipedia.org/wiki/Atoyac_(Veracruz)#Toponimia" rel="noreferrer nofollow">es.wikipedia.org/wiki/Atoyac_(Veracruz)#Toponimia</a>\n'
            "you'll find its coordinates"
        )

        assert fix_comment_text(comment_text) == expected_text

    @pytest.mark.vcr()
    def test_fixes_wikimedia_commons_link(self) -> None:
        """
        If there's a Wikimedia Commons URL with extra stuff in the suffix,
        that gets added to the URL.
        """
        # This example comes from Peter D. Tillman's comment:
        # https://www.flickr.com/photos/aahs_archives/23586820864/#comment72157667048114519
        #
        # Retrieved 22 January 2025
        comment_text = (
            'In use at <a href="https://en.wikipedia.org/wiki/Bell_YFM-1_Airacuda" rel="nofollow">en.wikipedia.org/wiki/Bell_YFM-1_Airacuda</a>\n'
            "From SDASM, Daniels collection:\n"
            '<a href="https://commons.wikimedia.org/wiki/File:Airacuda_Bell_XFM-1_" rel="nofollow">commons.wikimedia.org/wiki/File:Airacuda_Bell_XFM-1_</a>(15954491367).jpg\n'
            "Thanks for posting these!"
        )
        expected_text = (
            'In use at <a href="https://en.wikipedia.org/wiki/Bell_YFM-1_Airacuda" rel="nofollow">en.wikipedia.org/wiki/Bell_YFM-1_Airacuda</a>\n'
            "From SDASM, Daniels collection:\n"
            '<a href="https://commons.wikimedia.org/wiki/File:Airacuda_Bell_XFM-1_(15954491367).jpg" rel="noreferrer nofollow">commons.wikimedia.org/wiki/File:Airacuda_Bell_XFM-1_(15954491367).jpg</a>\n'
            "Thanks for posting these!"
        )

        assert fix_comment_text(comment_text) == expected_text

    @pytest.mark.vcr()
    def test_skips_correct_link(self) -> None:
        """
        A Wikipedia link that is correct is unchanged.
        """
        comment_text = (
            '<a href="https://en.wikipedia.org/wiki/Flickr" '
            'rel="noreferrer nofollow">en.wikipedia.org/wiki/Flickr</a>.'
        )

        assert fix_comment_text(comment_text) == comment_text

    @pytest.mark.vcr()
    def test_skips_unwanted_trailing_period(self) -> None:
        """
        If the trailing period isn't part of the Wikipedia page title,
        that period isn't added to the URL.
        """
        comment_text = 'You’re thinking of <a href="https://en.wikipedia.org/wiki/Longitude">en.wikipedia.org/wiki/Longitude</a>.'

        assert fix_comment_text(comment_text) == comment_text

    @pytest.mark.vcr()
    def test_skips_ambiguous_url(self) -> None:
        """
        If there's no Wikipedia page with or without the punctuation,
        it leaves the link as-is -- there's no obviously right thing
        for us to do here.
        """
        comment_text = 'This page <a href="https://en.wikipedia.org/wiki/DoesNotExist" rel="noreferrer nofollow">en.wikipedia.org/wiki/DoesNotExist</a>.'

        assert fix_comment_text(comment_text) == comment_text

    @pytest.mark.vcr()
    def test_skips_link_followed_by_whitespace(self) -> None:
        """
        A Wikipedia link followed by whitespace is unchanged.
        """
        comment_text = (
            '<a href="https://en.wikipedia.org/wiki/Flickr" '
            'rel="noreferrer nofollow">en.wikipedia.org/wiki/Flickr</a>\n'
            "That’s it, that’s the whole link."
        )

        assert fix_comment_text(comment_text) == comment_text

    @pytest.mark.vcr()
    def test_skips_homepage_link(self) -> None:
        """
        A link to the Wikipedia homepage is unchanged.
        """
        comment_text = (
            "This is the Wikipedia homepage: "
            '<a href="https://en.wikipedia.org/" '
            'rel="noreferrer nofollow">en.wikipedia.org/</a>.'
        )

        assert fix_comment_text(comment_text) == comment_text

    @pytest.mark.vcr()
    def test_skips_unrecognised_wikipedia(self) -> None:
        """
        A link to a language-specific Wikipedia that doesn't exist
        is skipped.
        """
        comment_text = (
            "This page doesn’t exist: "
            '<a href="https://zz.wikipedia.org/wiki/MadeUpPage" '
            'rel="noreferrer nofollow">zz.wikipedia.org/wiki/MadeUpPage</a>.'
        )

        assert fix_comment_text(comment_text) == comment_text



"""
Code for dealing with Flickr comments.

In particular, some of Flickr's auto-link detection can be a bit wonky,
e.g. it detects Wikipedia URLs at the wrong point, so we have some code
that fixes it.  This module allows us to apply similar fixes across
all of our projects.

Callers only need one function from this file: ``fix_comment_text``.
"""

from collections.abc import Iterator
import re
import typing
from xml.etree import ElementTree as ET

import httpx
import hyperlink

from .xml import find_required_elem


def fix_comment_text(comment_text: str) -> str:
    """
    Fix broken links and other markup issues in comment text.
    """
    try:
        comment_text = fix_wikipedia_links(comment_text)
    except Exception:
        pass

    return comment_text


class Link(typing.TypedDict):
    """
    Represents an <a> tag found in a block of text.
    """

    raw_markup: str
    href: str
    display_text: str
    suffix: str | None


def find_links(comment_text: str) -> Iterator[Link]:
    """
    Find links in a block of text.

    For this function, we rely on the fact that Flickr auto-links in
    a very particular way, for example:

        <a href="https://example.com" rel="noreferrer nofollow">example.com</a>

    It might be safer to use a proper parser like BeautifulSoup,
    but a Flickr comment isn't a complete HTML document and using
    regex makes it easier to do minimally invasive replacements.

    Note: some older comments omit the ``noreferrer`` attribute.

    """
    # This regex is designed to match all <a> tags in the form above,
    # then capture everything after that up to the next bit of whitespace.
    for m in re.finditer(
        r'<a href="(?P<href>[^"]+)" '
        r'rel="(?:noreferrer )?nofollow">'
        r"(?P<display_text>[^<]+)"
        r"</a>"
        r"(?P<suffix>[^\s]*)",
        comment_text,
    ):
        yield {
            "raw_markup": m.group(0),
            "href": m.group("href"),
            "display_text": m.group("display_text"),
            "suffix": m.group("suffix") or None,
        }


def fix_wikipedia_links(comment_text: str) -> str:
    """
    Fix Wikipedia links in Flickr comments.

    Flickr's comment parser will try to auto-detect links, and it assumes
    punctuation isn't part of links, but this breaks links to some
    Wikipedia pages, for example:

        https://en.wikipedia.org/wiki/William_Barnes_Jr.

    It will omit the final period from the link, which means it goes to
    the wrong page.

    This is a particular issue for Flickr Commons, where people leave
    a lot of comments linking to Wikipedia pages.

    This function will fix the Wikipedia links auto-detected by Flickr.
    It moves any trailing punctuation that's part of the link inside
    the <a>.  We should never change the text of the comment, just move
    bits in/out of the <a>.

    See https://github.com/Flickr-Foundation/commons.flickr.org/issues/297
    """
    for link in find_links(comment_text):
        # If this link is immediately followed by whitespace, we can be
        # confident we've got the full link.
        if link["suffix"] is None:
            continue

        # Parse the URL and get the title of the page
        #
        # We're looking for a URL of the form
        #
        #     https://en.wikipedia.org/wiki/{title}
        #
        # Anything else we should skip.
        url = hyperlink.parse(link["href"])

        is_english_wikipedia = url.host in {
            "en.wikipedia.org",
            "en.m.wikipedia.org",
            "commons.wikimedia.org",
        }

        # It's easier to write a regex than enumerate all of these.
        # e.g. de.wikipedia.org, fr.wikipedia.org
        is_other_wikipedia = re.match(r"^[a-z]{2}\.wikipedia.org$", url.host)

        if not is_english_wikipedia and not is_other_wikipedia:
            continue

        if len(url.path) < 2 or url.path[0] != "wiki":
            continue

        orig_title = url.path[1]

        # If there's a Wikipedia page with this exact title, then the
        # link works and we can leave it as-is.
        if _get_wikipedia_page(host=url.host, title=orig_title) == "found":
            continue

        # Otherwise, check to see if there's a page with the suffix
        # added -- and if there does, use that as the new link.
        #
        # We don't include the fragment when looking up the title of
        # the page, but we do add it if we're fixing the URL.
        if "#" in link["suffix"]:
            suffix, fragment = link["suffix"].split("#")
            fragment = f"#{fragment}"
        else:
            suffix = link["suffix"]
            fragment = ""

        alt_title = orig_title + suffix

        if _get_wikipedia_page(host=url.host, title=alt_title) == "found":
            new_url = f"{url.host}/wiki/{alt_title}{fragment}"

            comment_text = comment_text.replace(
                link["raw_markup"],
                (
                    f'<a href="{url.scheme}://{new_url}" '
                    'rel="noreferrer nofollow">'
                    f"{new_url}</a>"
                ),
            )

    return comment_text


WikipediaPageStatus = typing.Literal["found", "redirected", "not_found"]


def _get_wikipedia_page(host: str, title: str) -> WikipediaPageStatus:
    """
    Look up a page on Wikipedia and see whether it:

    1.  Exists, with the given title
    2.  Exists, but the title is normalized/redirected
    3.  Isn't found

    """
    if host == "en.m.wikipedia.org":
        host = "en.wikipedia.org"

    resp = httpx.get(
        f"https://{host}/w/api.php",
        params={
            "action": "query",
            "prop": "revisions",
            "titles": title,
            "rvprop": "timestamp",
            "format": "xml",
        },
    )
    resp.raise_for_status()

    # Note: the ElementTree API is not hardened against untrusted XML,
    # but we trust the Wikipedia API enough to try this.
    xml = ET.fromstring(resp.text)

    # The API response will contain a single ``page`` element,
    # like so:
    #
    #    <?xml version="1.0"?>
    #    <api batchcomplete="">
    #      <query>
    #        <normalized>
    #          <n from="William_Barnes_Jr" to="William Barnes Jr"/>
    #        </normalized>
    #        <pages>
    #          <page _idx="-1" ns="0" title="William Barnes Jr" missing=""/>
    #        </pages>
    #      </query>
    #    </api>

    # If the <page> element has the ``missing`` attribute, then there's
    # no such page.
    page = find_required_elem(xml, path=".//page")

    if "missing" in page.attrib:
        return "not_found"

    # If the <page> element has the exact same title as the thing we
    # searched for, then it exists.
    #
    # (Note: Wikipedia replaces spaces with underscores in URLs, which
    # we undo to compare titles.)
    if page.attrib["title"] == title.replace("_", " "):
        return "found"

    # If we found a <page> element but it doesn't have the expected title,
    # we may have been redirected.
    if xml.find(".//normalized") is not None:
        return "redirected"

    # This should never happen in practice so we can't test it, but we
    # include it for the sake of easy debugging if it does.
    else:  # pragma: no cover
        raise RuntimeError(
            f"Unable to parse Wikipedia API response for {title!r}: {resp.text}"
        )



"""
Allow encoding/decoding JSON in a type-preserving way.

In particular, these classes allow us to store Python-native values
like datetime, date, and Path in JSON and retrieve them as nicely typed
Python objects, not just strings.
"""

import datetime
import json
import pathlib
import typing


T = typing.TypeVar("T")


class NitrateEncoder(json.JSONEncoder):
    """
    A custom JSON encoder that supports datetimes, dates and paths.

        >>> t = datetime.datetime(2001, 2, 3, 4, 5, 6)
        >>> json.dumps({"t": t}, cls=NitrateEncoder)
        '{"t": {"type": "datetime.datetime", "value": "2001-02-03T04:05:06"}}'

    This is meant to be used with ``NitrateDecoder`` -- together, they
    allow you to serialise a datetime value via JSON and preserve its type.

    """

    def default(self, t: T) -> typing.Any:
        """
        Convert a Python value to a JSON value.
        """
        if isinstance(t, datetime.datetime):
            return {"type": "datetime.datetime", "value": t.isoformat()}
        elif isinstance(t, datetime.date):
            return {"type": "datetime.date", "value": t.isoformat()}
        elif isinstance(t, pathlib.Path):
            return {"type": "pathlib.Path", "value": str(t)}
        else:
            return super().default(t)


class NitrateDecoder(json.JSONDecoder):
    """
    A custom JSON decoder that supports the datetimes encoded
    by NitrateEncoder.

        >>> json.loads(
        ...     '{"t": {"type": "datetime.datetime", "value": "2001-02-03T04:05:06"}}',
        ...     cls=NitrateDecoder)
        {'t': datetime.datetime(2001, 2, 3, 4, 5, 6)}

    """

    def __init__(self) -> None:
        """
        Create a new JSONDecoder.

        The ``object_hook`` will be called with the result of any
        object literal that gets decoded.
        """
        super().__init__(object_hook=self.dict_to_object)

    def dict_to_object(self, d: dict[str, typing.Any]) -> typing.Any:
        """
        Convert a JSON value to a Python-native value.
        """
        if d.keys() == {"type", "value"}:
            if d.get("type") == "datetime.datetime":
                return datetime.datetime.fromisoformat(d["value"])
            elif d.get("type") == "datetime.date":
                return datetime.datetime.fromisoformat(d["value"]).date()
            elif d.get("type") == "pathlib.Path":
                return pathlib.Path(d["value"])

        return d


__all__ = ["NitrateEncoder", "NitrateDecoder"]



"""
Tests for ``nitrate.json``.
"""

import datetime
import json
import pathlib
import typing
import uuid

import pytest

from nitrate.json import NitrateDecoder, NitrateEncoder


@pytest.mark.parametrize(
    "value",
    [
        "an interesting time",
        1234,
        None,
        ["1", "2", "3"],
        {"sides": 5, "color": "red"},
        datetime.datetime(2001, 2, 3, 4, 5, 6),
        datetime.date(2001, 2, 3),
        pathlib.Path("x/y/z"),
    ],
)
def test_can_json_round_trip(value: typing.Any) -> None:
    """
    If you encode JSON with ``NitrateEncoder``, you can decode it
    with ``NitrateDecoder`` and get the same value back.
    """
    json_string = json.dumps(value, cls=NitrateEncoder)
    parsed_json_value = json.loads(json_string, cls=NitrateDecoder)

    assert parsed_json_value == value


@pytest.mark.parametrize(
    ["json_string", "value"],
    [
        (
            '{"type": "datetime.datetime", "value": "2001-02-03T04:05:06"}',
            datetime.datetime(2001, 2, 3, 4, 5, 6),
        ),
        (
            '{"type": "datetime.date", "value": "2001-02-03"}',
            datetime.date(2001, 2, 3),
        ),
        (
            '{"type": "pathlib.Path", "value": "x/y/z"}',
            pathlib.Path("x/y/z"),
        ),
    ],
)
def test_can_decode_json(json_string: str, value: typing.Any) -> None:
    """
    We can decode JSON strings created with ``NitrateEncoder``.

    Note: we use hard-coded strings here to ensure the decoder remains
    compatible across versions, e.g. something encoded with nitrate 1.1
    can be decoded with nitrate 1.2.
    """
    assert json.loads(json_string, cls=NitrateDecoder) == value


def test_encoding_an_unrecognised_type_still_fails() -> None:
    """
    Trying to encode an unrecognised type will fail, even if you use
    ``NitrateEncoder``.

    This is a regression test for an old bug where ``NitrateEncoder``
    would silently drop values with unrecognised types.
    """
    with pytest.raises(TypeError, match="Object of type UUID is not JSON serializable"):
        json.dumps({"id": uuid.uuid4()})

    with pytest.raises(TypeError, match="Object of type UUID is not JSON serializable"):
        json.dumps({"id": uuid.uuid4()}, cls=NitrateEncoder)


@pytest.mark.parametrize(
    ["json_string", "value"],
    [
        (
            '{"type": "datetime.date", "string": "2001-01-01"}',
            {"type": "datetime.date", "string": "2001-01-01"},
        ),
        (
            '{"type": "datetime.date", "value": "2001-01-01", "day": "monday"}',
            {"type": "datetime.date", "value": "2001-01-01", "day": "monday"},
        ),
        (
            '{"type": "Currency", "value": "£123.45"}',
            {"type": "Currency", "value": "£123.45"},
        ),
    ],
)
def test_decoding_an_unrecognised_type_is_okay(
    json_string: str, value: typing.Any
) -> None:
    """
    Trying to decode a JSON object which includes a ``type`` key will
    leave it as-is, if it doesn't look like it was encoded by nitrate.
    """
    assert json.loads(json_string, cls=NitrateDecoder) == value
